# coding: utf-8
import collections
import itertools
import logging
from typing import Dict, Iterator, List, Tuple

import cached_property
import pulp

from haoda import ir, util

_logger = logging.getLogger().getChild(__name__)


class SuperSourceNode(ir.Module):
  """A node representing the super source in the dataflow graph.

  A super source doesn't have parent nodes.

  Attributes:
      fwd_nodes (Dict[Tuple[str, int], ForwardNode]): Dict mapping tuples of
        (tensor name, offset) to nodes.
      cpt_nodes (Dict[Tuple[str, int], ComputeNode]): Dict mapping tuples of
        (stage_name, pe_id) to nodes.
      super_sink (SuperSinkNode): The super sink node of this DAG.
  """

  def __init__(
      self,
      fwd_nodes: Dict[Tuple[str, int], 'ForwardNode'],
      cpt_nodes: Dict[Tuple[str, int], 'ComputeNode'],
      super_sink: 'SuperSinkNode',
  ):
    super().__init__()
    self.fwd_nodes = fwd_nodes
    self.cpt_nodes = cpt_nodes
    self.super_sink = super_sink

  @property
  def graphviz(self) -> str:
    output = 'digraph {\n'
    for src, dst in self.dfs_edge_gen():
      output += f'  "{repr(src)}" -> "{repr(dst)}"\n'
    output += '}\n'
    return output

  def verify_mode_depths(self) -> None:
    latency_table = {}
    lp_problem = pulp.LpProblem('verify_fifo_depths', pulp.LpMinimize)
    for node in self.tpo_valid_node_gen():
      if self in node.parents:
        latency_table[node] = 0
      else:
        latency_table[node] = pulp.LpVariable(
            name=f'latency_{node.name}',
            lowBound=0,
            cat='Integer',
        )
        lp_problem.extend(
            parent.get_latency(node) +
            latency_table[parent] <= latency_table[node]
            for parent in node.parents)
        lp_problem.extend(
            parent.get_latency(node) + latency_table[parent] +
            parent.fifo(node).depth >= latency_table[node]
            for parent in node.parents)

    lp_status = lp_problem.solve()
    if lp_status == pulp.LpStatusOptimal:
      _logger.debug('II=1 check: PASS')
    elif lp_status == pulp.LpStatusInfeasible:
      _logger.warn('II=1 check: FAIL')
    else:
      lp_status_str = pulp.LpStatus[lp_status]
      _logger.error('ILP error: %s\n%s', lp_status_str, lp_problem)
      raise util.InternalError('unexpected ILP status: %s' % lp_status_str)

    for node in self.tpo_valid_node_gen():
      if self in node.parents:
        min_capacity = 0
      else:
        min_capacity = min(
            parent.get_latency(node) + int(pulp.value(latency_table[parent])) +
            parent.fifo(node).depth for parent in node.parents)

      debug_enabled = _logger.isEnabledFor(logging.DEBUG)
      check_failed = int(pulp.value(latency_table[node])) > min_capacity
      if debug_enabled or check_failed:
        (_logger.debug if debug_enabled else _logger.warn)(
            'II=1 check %s: %s: latency %d %s min capacity %d',
            '✖' if check_failed else '✔',
            repr(node),
            int(pulp.value(latency_table[node])),
            '>' if check_failed else '<=',
            min_capacity,
        )

  def update_module_depths(
      self,
      depths: Dict[int, int],
  ) -> None:
    """Update module pipeline depths and FIFO depths.

    The FIFO depths are determined by solving an ILP problem:

    + Optimization objective: minimize the sum (weighted by FIFO width) of all
    FIFO depths.
    + Constraints: the whole DAG can be fully pipelined without artificial
    stalls.

    For every non-overlapping path between a pair of nodes,
    the latency of each token is the maximum minimum latency among all paths.
    To enable full pipelining,
    this latency must not exceed the maximum latency of any path.
    The minimum latency of each path is the sum of the FIFO write latency in
    each module and the number of edges (FIFOs),
    since the minimum latency of a FIFO is 1.
    The maximum latency of each path is the sum of the FIFO write latency in
    each module and the total depth of FIFOs.

    Args:
        depths (Dict[int, int]): Dict mapping module ids to pipeline depths.
    """
    # update module pipeline depths
    for src_node, dst_node in self.bfs_valid_edge_gen():
      module_id = self.module_table[src_node][1]
      depth = depths.get(module_id)
      if depth is not None:
        fifo = src_node.fifo(dst_node)
        if fifo.write_lat != depth:
          _logger.debug('%s write latency changed %s -> %d', fifo,
                        fifo.write_lat, depth)
          fifo.write_lat = depth

    # set up ILP problem, variables, and objective
    lp_problem = pulp.LpProblem('optimal_fifo_depths', pulp.LpMinimize)
    lp_vars = {}
    for src_node, dst_node in self.bfs_valid_edge_gen():
      lp_vars[(src_node, dst_node)] = pulp.LpVariable(
          name=f'depth_{src_node.fifo(dst_node).c_expr}',
          lowBound=0,
          cat='Integer',
      )
    lp_problem += sum(
        x.fifo(y).haoda_type.width_in_bits * v for (x, y), v in lp_vars.items())

    # add ILP constraints
    latency_table = {
        x: pulp.LpVariable(name=f'latency_{x.name}', lowBound=0, cat='Integer')
        for x in self.tpo_valid_node_gen()
    }
    for node in self.tpo_valid_node_gen():
      if self in node.parents:
        latency_table[node] = 0
      else:
        lp_problem.extend(
            parent.get_latency(node) +
            latency_table[parent] <= latency_table[node]
            for parent in node.parents)
        lp_problem.extend(
            parent.get_latency(node) + latency_table[parent] +
            lp_vars[(parent, node)] >= latency_table[node]
            for parent in node.parents)

    # solve ILP
    lp_status = lp_problem.solve()
    if lp_status != pulp.LpStatusOptimal:
      lp_status_str = pulp.LpStatus[lp_status]
      _logger.error('ILP error: %s\n%s', lp_status_str, lp_problem)
      raise util.InternalError('unexpected ILP status: %s' % lp_status_str)

    # update FIFO depths
    for (src_node, dst_node), lp_var in lp_vars.items():
      depth = int(pulp.value(lp_var))
      fifo = src_node.fifo(dst_node)
      if fifo.depth != depth:
        _logger.debug('%s * depth %d -> %d', fifo, fifo.depth, depth)
        fifo.depth = depth

    self.verify_mode_depths()

  @property
  def name(self):
    return 'super_source'

  def __repr__(self) -> str:
    return '\033[35msuper source\033[0m'

  @cached_property.cached_property
  def module_table(self) -> Dict[ir.Node, Tuple[ir.ModuleTrait, int]]:
    """Module table maps an IR node to (module_trait, module_id).

    Returns:
      A dict mapping an IR node to (module_trait, module_id) tuple.
    """
    self._module_traits: Dict[ir.ModuleTrait,
                              List[ir.Node]] = collections.OrderedDict()
    module_table: Dict[ir.Node, Tuple[ir.ModuleTrait,
                                      int]] = collections.OrderedDict()

    for node in self.tpo_valid_node_gen():
      self._module_traits.setdefault(ir.ModuleTrait(node), []).append(node)
    for idx, module_trait in enumerate(self._module_traits):
      for node in self._module_traits[module_trait]:
        module_table[node] = module_trait, idx
    return module_table

  @cached_property.cached_property
  def module_traits(self) -> Tuple[ir.ModuleTrait, ...]:
    return tuple(self.module_trait_table)

  @property
  def module_trait_table(self) -> Dict[ir.ModuleTrait, List[ir.Node]]:
    # pylint: disable=pointless-statement
    self.module_table
    return self._module_traits

  def tpo_valid_node_gen(self) -> Iterator[ir.Module]:
    """Traverse valid descendant nodes in tpological order.

    Load and store nodes are ordered in the same way as they are specified in
    soda files.

    Yields:
        Iterator[ir.Module]: Nodes that are not super source or super sink.
    """
    yield from self.load_nodes
    yield from filter(
        lambda x: not isinstance(x, MemoryNode) and is_valid_node(x),
        super().tpo_node_gen(),
    )
    yield from self.store_nodes

  def bfs_valid_edge_gen(self) -> Iterator[ir.Module]:
    return filter(is_valid_edge, self.bfs_edge_gen())

  @property
  def load_nodes(self) -> Tuple['LoadNode', ...]:
    return self.children

  @property
  def store_nodes(self) -> Tuple['StoreNode', ...]:
    return self.super_sink.parents


class SuperSinkNode(ir.Module):
  """A node representing the super sink in the dataflow graph.

  A super sink doesn't have child nodes.
  """

  @property
  def name(self):
    return 'super_sink'

  def __repr__(self) -> str:
    return '\033[34msuper sink\033[0m'


class ForwardNode(ir.Module):
  """A node representing a forward module in the dataflow graph.

  Attributes:
    tensor: Tensor corresponding to this node.
    offset: Int representing the offset of this tensor.
  """

  def __init__(self, **kwargs):
    super().__init__()
    self.tensor = kwargs.pop('tensor')
    self.offset = kwargs.pop('offset')

  def __repr__(self):
    return '\033[32mforward %s @%d\033[0m' % (self.tensor.name, self.offset)

  @property
  def name(self):
    return '{}_offset_{}'.format(self.tensor.name, self.offset)


class ComputeNode(ir.Module):
  """A node representing a compute module in the dataflow graph.

  Attributes:
    tensor: Tensor corresponding to this node.
    pe_id: Int representing the PE id.
    fifo_map: {str: {idx: Node}}
  """

  def __init__(self, **kwargs):
    super().__init__()
    self.tensor = kwargs.pop('tensor')
    self.pe_id = kwargs.pop('pe_id')
    self.fifo_map = collections.defaultdict(dict)

  def __repr__(self):
    return '\033[31mcompute %s #%d\033[0m' % (self.tensor.name, self.pe_id)

  @property
  def name(self):
    return '{}_pe_{}'.format(self.tensor.name, self.pe_id)


class MemoryNode(ir.Module):

  def __init__(self, var: str, bank: int):
    super().__init__()
    self.var = var
    self.bank = bank

  @property
  def name(self) -> str:
    return f'{self.var}_bank_{self.bank}'

  def __str__(self) -> str:
    return f'dram<bank {self.bank} {self.var}>'


class LoadNode(MemoryNode):

  def __repr__(self) -> str:
    return f'\033[33mload {self.var} bank {self.bank}\033[0m'


class StoreNode(MemoryNode):

  def __repr__(self) -> str:
    return f'\033[36mstore {self.var} bank {self.bank}\033[0m'


def is_valid_node(node: ir.Module) -> bool:
  return not isinstance(node, (SuperSourceNode, SuperSinkNode))


def is_valid_edge(edge: Tuple[ir.Module, ir.Module]) -> bool:
  return all(map(is_valid_node, edge))


# pylint: disable=too-many-branches,too-many-statements
def create_dataflow_graph(stencil):
  chronological_tensors = stencil.chronological_tensors
  super_source = SuperSourceNode(
      fwd_nodes={},
      cpt_nodes={},
      super_sink=SuperSinkNode(),
  )

  load_nodes = {
      stmt.name:
      tuple(LoadNode(var=stmt.name, bank=bank) for bank in stmt.dram)
      for stmt in stencil.input_stmts
  }
  store_nodes = {
      stmt.name:
      tuple(StoreNode(var=stmt.name, bank=bank) for bank in stmt.dram)
      for stmt in stencil.output_stmts
  }

  for mem_node in itertools.chain(*load_nodes.values()):
    super_source.add_child(mem_node)
  for mem_node in itertools.chain(*store_nodes.values()):
    mem_node.add_child(super_source.super_sink)

  def color_id(node):
    if isinstance(node, LoadNode):
      return f'\033[33mload {node.var}[bank{node.bank}]\033[0m'
    if isinstance(node, StoreNode):
      return f'\033[36mstore {node.var}[bank{node.bank}]\033[0m'
    if isinstance(node, ForwardNode):
      return f'\033[32mforward {node.tensor.name} @{node.offset}\033[0m'
    if isinstance(node, ComputeNode):
      return f'\033[31mcompute {node.tensor.name} #{node.pe_id}\033[0m'
    return 'unknown node'

  def color_attr(node):
    result = []
    for k, v in node.__dict__.items():
      if (node.__class__, k) in ((SuperSourceNode, 'parents'), (SuperSinkNode,
                                                                'children')):
        continue
      if k in ('parents', 'children'):
        result.append('%s: [%s]' % (k, ', '.join(map(color_id, v))))
      else:
        result.append('%s: %s' % (k, repr(v)))
    return '{%s}' % ', '.join(result)

  def color_print(node):
    return '%s: %s' % (color_id(node), color_attr(node))

  print_node = color_id

  if stencil.replication_factor > 1:
    replicated_next_fifo = stencil.get_replicated_next_fifo()
    replicated_all_points = stencil.get_replicated_all_points()
    replicated_reuse_buffers = stencil.get_replicated_reuse_buffers()

    def add_fwd_nodes(src_name):
      dsts = replicated_all_points[src_name]
      reuse_buffer = replicated_reuse_buffers[src_name][1:]
      nodes_to_add = []
      for dst_point_dicts in dsts.values():
        for offset in dst_point_dicts:
          if (src_name, offset) in super_source.fwd_nodes:
            continue
          fwd_node = ForwardNode(
              tensor=stencil.tensors[src_name],
              offset=offset,
              depth=stencil.get_replicated_reuse_buffer_length(
                  src_name, offset))
          _logger.debug('create %s', print_node(fwd_node))
          init_offsets = [start for start, end in reuse_buffer if start == end]
          if offset in init_offsets:
            if src_name in [stencil.input.name]:
              load_node_count = len(load_nodes[src_name])
              load_nodes[src_name][load_node_count - 1 -
                                   offset % load_node_count].add_child(fwd_node)
            else:
              (super_source.cpt_nodes[(src_name, 0)].add_child(fwd_node))
          super_source.fwd_nodes[(src_name, offset)] = fwd_node
          if offset in replicated_next_fifo[src_name]:
            nodes_to_add.append(
                (fwd_node, (src_name, replicated_next_fifo[src_name][offset])))
      for src_node, key in nodes_to_add:
        src_node.add_child(super_source.fwd_nodes[key])

    add_fwd_nodes(stencil.input.name)

    for stage in stencil.get_stages_chronologically():
      cpt_node = ComputeNode(stage=stage, pe_id=0)
      _logger.debug('create %s', print_node(cpt_node))
      super_source.cpt_nodes[(stage.name, 0)] = cpt_node
      for input_name, input_window in stage.window.items():
        for i in range(len(input_window)):
          offset = next(offset for offset, points in (
              replicated_all_points[input_name][stage.name].items())
                        if points == i)
          fwd_node = super_source.fwd_nodes[(input_name, offset)]
          _logger.debug('  access %s', print_node(fwd_node))
          fwd_node.add_child(cpt_node)
      if stage.is_output():
        super_source.cpt_nodes[stage.name,
                               0].add_child(store_nodes[stage.name][0])
      else:
        add_fwd_nodes(stage.name)

  else:
    next_fifo = stencil.next_fifo
    all_points = stencil.all_points
    reuse_buffers = stencil.reuse_buffers

    def add_fwd_nodes(src_name):
      dsts = all_points[src_name]
      reuse_buffer = reuse_buffers[src_name][1:]
      nodes_to_add = []
      for dst_point_dicts in dsts.values():
        for offset in dst_point_dicts:
          if (src_name, offset) in super_source.fwd_nodes:
            continue
          fwd_node = ForwardNode(tensor=stencil.tensors[src_name],
                                 offset=offset)
          #depth=stencil.get_reuse_buffer_length(src_name, offset))
          _logger.debug('create %s', print_node(fwd_node))
          # init_offsets is the start of each reuse chain
          init_offsets = [
              next(end
                   for start, end in reuse_buffer
                   if start == unroll_idx)
              for unroll_idx in reversed(range(stencil.unroll_factor))
          ]
          _logger.debug('reuse buffer: %s', reuse_buffer)
          _logger.debug('init offsets: %s', init_offsets)
          if offset in init_offsets:
            if src_name in stencil.input_names:
              # fwd from external input
              load_node_count = len(load_nodes[src_name])
              load_nodes[src_name][load_node_count - 1 -
                                   offset % load_node_count].add_child(fwd_node)
            else:
              # fwd from output of last stage
              # tensor name and offset are used to find the cpt node
              cpt_offset = next(
                  unroll_idx for unroll_idx in range(stencil.unroll_factor)
                  if init_offsets[unroll_idx] == offset)
              cpt_node = super_source.cpt_nodes[(src_name, cpt_offset)]
              cpt_node.add_child(fwd_node)
          super_source.fwd_nodes[(src_name, offset)] = fwd_node
          if offset in next_fifo[src_name]:
            nodes_to_add.append(
                (fwd_node, (src_name, next_fifo[src_name][offset])))
      for src_node, key in nodes_to_add:
        # fwd from another fwd node
        src_node.add_child(super_source.fwd_nodes[key])

    for input_name in stencil.input_names:
      add_fwd_nodes(input_name)

    for tensor in chronological_tensors:
      if tensor.is_input():
        continue
      for unroll_index in range(stencil.unroll_factor):
        pe_id = stencil.unroll_factor - 1 - unroll_index
        cpt_node = ComputeNode(tensor=tensor, pe_id=pe_id)
        _logger.debug('create %s', print_node(cpt_node))
        super_source.cpt_nodes[(tensor.name, pe_id)] = cpt_node
        for input_name, input_window in tensor.ld_indices.items():
          for i in range(len(input_window)):
            offset = next(offset for offset, points in all_points[input_name][
                tensor.name].items() if pe_id in points and points[pe_id] == i)
            fwd_node = super_source.fwd_nodes[(input_name, offset)]
            _logger.debug('  access %s', print_node(fwd_node))
            fwd_node.add_child(cpt_node)
      if tensor.is_output():
        for pe_id in range(stencil.unroll_factor):
          super_source.cpt_nodes[tensor.name, pe_id].add_child(
              store_nodes[tensor.name][pe_id % len(store_nodes[tensor.name])])
      else:
        add_fwd_nodes(tensor.name)

  # pylint: disable=too-many-nested-blocks
  for src_node in super_source.tpo_valid_node_gen():
    for dst_node in filter(is_valid_node, src_node.children):
      # 5 possible edge types:
      # 1. load => fwd
      # 2. fwd => fwd
      # 3. fwd => cpt
      # 4. cpt => fwd
      # 5. cpt => store
      if isinstance(src_node, LoadNode):
        write_lat = 0
      elif isinstance(src_node, ForwardNode):
        write_lat = 2
      elif isinstance(src_node, ComputeNode):
        write_lat = src_node.tensor.st_ref.lat
      else:
        raise util.InternalError('unexpected source node: %s' % repr(src_node))

      fifo = ir.FIFO(src_node, dst_node, depth=0, write_lat=write_lat)
      lets: List[ir.Let] = []
      if isinstance(src_node, LoadNode):
        expr = ir.DRAMRef(
            haoda_type=dst_node.tensor.haoda_type,
            dram=(src_node.bank,),
            var=dst_node.tensor.name,
            offset=(stencil.unroll_factor - 1 - dst_node.offset) //
            len(stencil.stmt_table[dst_node.tensor.name].dram),
        )
      elif isinstance(src_node, ForwardNode):
        if isinstance(dst_node, ComputeNode):
          dst = src_node.tensor.children[dst_node.tensor.name]
          src_name = src_node.tensor.name
          unroll_idx = dst_node.pe_id
          point = all_points[src_name][dst.name][src_node.offset][unroll_idx]
          idx = list(dst.ld_indices[src_name].values())[point].idx
          _logger.debug('%s%s referenced by <%s> @ unroll_idx=%d is %s',
                        src_name, util.idx2str(idx), dst.name, unroll_idx,
                        print_node(src_node))
          dst_node.fifo_map[src_name][idx] = fifo
        delay = stencil.reuse_buffer_lengths[src_node.tensor.name]\
                                            [src_node.offset]
        offset = src_node.offset - delay
        for parent in src_node.parents:  # fwd node has only 1 parent
          for fifo_r in parent.fifos:
            if fifo_r.edge == (parent, src_node):
              break
        if delay > 0:
          # TODO: build an index somewhere
          for let in src_node.lets:
            # pylint: disable=undefined-loop-variable
            if isinstance(let.expr, ir.DelayedRef) and let.expr.ref == fifo_r:
              var_name = let.name
              var_type = let.haoda_type
              break
          else:
            var_name = 'let_%d' % len(src_node.lets)
            # pylint: disable=undefined-loop-variable
            var_type = fifo_r.haoda_type
            lets.append(
                ir.Let(haoda_type=var_type,
                       name=var_name,
                       expr=ir.DelayedRef(delay=delay, ref=fifo_r)))
          expr = ir.Var(name=var_name, idx=[])
          expr.haoda_type = var_type
        else:
          expr = fifo_r  # pylint: disable=undefined-loop-variable
      elif isinstance(src_node, ComputeNode):

        def replace_refs_callback(obj, args):
          if isinstance(obj, ir.Ref):
            _logger.debug(
                'replace %s with %s',
                obj,
                # pylint: disable=cell-var-from-loop,undefined-loop-variable
                src_node.fifo_map[obj.name][obj.idx])
            # pylint: disable=cell-var-from-loop,undefined-loop-variable
            return src_node.fifo_map[obj.name][obj.idx]
          return obj

        _logger.debug('lets: %s', src_node.tensor.lets)
        lets = [_.visit(replace_refs_callback) for _ in src_node.tensor.lets]
        _logger.debug('replaced lets: %s', lets)
        _logger.debug('expr: %s', src_node.tensor.expr)
        expr = src_node.tensor.expr.visit(replace_refs_callback)
        _logger.debug('replaced expr: %s', expr)
        if isinstance(dst_node, StoreNode):
          dram_ref = ir.DRAMRef(
              haoda_type=src_node.tensor.haoda_type,
              dram=(dst_node.bank,),
              var=src_node.tensor.name,
              offset=(src_node.pe_id) //
              len(stencil.stmt_table[src_node.tensor.name].dram),
          )
          dst_node.lets.append(ir.Let(haoda_type=None, name=dram_ref,
                                      expr=fifo))
      else:
        raise util.InternalError('unexpected node of type %s' % type(src_node))

      src_node.exprs[fifo] = expr
      src_node.lets.extend(_ for _ in lets if _ not in src_node.lets)
      _logger.debug('fifo [%d]: %s%s => %s', fifo.depth, color_id(src_node),
                    '' if fifo.write_lat is None else ' ~%d' % fifo.write_lat,
                    color_id(dst_node))

  super_source.update_module_depths({})

  return super_source



import collections
import itertools
import logging
import operator
from typing import Dict, List, Tuple, Union

import cached_property
import pulp
import toposort
from haoda import ir, util
from haoda.ir import arithmetic

import soda.tensor
import soda.util
from soda import dataflow, visitor
from soda.optimization import computation_reuse as cr
from soda.optimization import inline

_logger = logging.getLogger().getChild(__name__)


class Stencil():
  """
  Attributes:
    iterate: int, number of iteration to implement.
    border: Reserved.
    preserve_border: Reserved.
    cluster: Reserved.
    burst_width: int, width of bits for DRAM burst access.
    app_name: str, application's name.
    tile_size: List of int.
    unroll_factor: int.
    replication_factor: int.
    optimizations: Set of enabled optimizations.
    dim: int.
    param_stmts: List of ParamStmt.
    input_stmts: List of InputStmt.
    local_stmts: List of LocalStmt.
    output_stmts: List of OutputStmt.

  Cached properties:
    tensors: Dict from str of name to Tensor.
    input_names: Tuple of str, names of input tensors.
    param_names: Tuple of str, names of param tensors.
    local_names: Tuple of str, names of local tensors.
    output_names: Tuple of str, names of output tensors.
  """

  def __init__(self, **kwargs):
    self.iterate = kwargs.pop('iterate')
    if self.iterate < 1:
      raise util.SemanticError('cannot iterate %d times' % self.iterate)
    self.border = kwargs.pop('border')
    self.preserve_border = self.border == 'preserve'
    self.cluster = kwargs.pop('cluster')
    # platform determined
    self.burst_width = kwargs.pop('burst_width')
    # application determined
    self.app_name = kwargs.pop('app_name')
    # parameters that can be explored
    self.tile_size = tuple(kwargs.pop('tile_size'))
    self.unroll_factor = kwargs.pop('unroll_factor')
    self.replication_factor = kwargs.pop('replication_factor')
    # stage-independent
    self.dim = kwargs.pop('dim')
    self.param_stmts = kwargs.pop('param_stmts')
    # stage-specific
    self.input_stmts = kwargs.pop('input_stmts')
    self.local_stmts = kwargs.pop('local_stmts')
    self.output_stmts = kwargs.pop('output_stmts')
    self.optimizations = {}
    if 'optimizations' in kwargs:
      self.optimizations = kwargs.pop('optimizations')

    if 'dram_in' in kwargs:
      dram_in = kwargs.pop('dram_in')
      if dram_in is not None:
        if ':' in dram_in:
          input_stmt_map = {_.name: _ for _ in self.input_stmts}
          for dram_map in dram_in.split('^'):
            var_name, bank_list = dram_map.split(':')
            if var_name not in input_stmt_map:
              raise util.SemanticError('no input named `{}`'.format(var_name))
            input_stmt_map[var_name].dram = tuple(map(int,
                                                      bank_list.split('.')))
        else:
          for input_stmt in self.input_stmts:
            input_stmt.dram = tuple(map(int, dram_in.split('.')))

    if 'dram_out' in kwargs:
      dram_out = kwargs.pop('dram_out')
      if dram_out is not None:
        if ':' in dram_out:
          output_stmt_map = {_.name: _ for _ in self.output_stmts}
          for dram_map in dram_out.split(','):
            var_name, bank_list = dram_map.split(':')
            if var_name not in output_stmt_map:
              raise util.SemanticError('no output named `{}`'.format(var_name))
            output_stmt_map[var_name].dram = tuple(
                map(int, bank_list.split('.')))
        else:
          for output_stmt in self.output_stmts:
            output_stmt.dram = tuple(map(int, dram_out.split('.')))

    if self.iterate > 1:
      if len(self.input_stmts) != len(self.output_stmts):
        raise util.SemanticError(
            'number of input tensors must be the same as output if iterate > 1 '
            'times, currently there are %d input(s) but %d output(s)' %
            (len(self.input_stmts), len(self.output_stmts)))
      if self.input_types != self.output_types:
        raise util.SemanticError(
            'input must have the same type(s) as output if iterate > 1 '
            'times, current input has type %s but output has type %s' %
            (util.lst2str(self.input_types), util.lst2str(self.output_types)))
      _logger.debug(
          'pipeline %d iterations of [%s] -> [%s]' %
          (self.iterate, ', '.join('%s: %s' % (stmt.haoda_type, stmt.name)
                                   for stmt in self.input_stmts), ', '.join(
                                       '%s: %s' % (stmt.haoda_type, stmt.name)
                                       for stmt in self.output_stmts)))

    for stmt in itertools.chain(self.local_stmts, self.output_stmts):
      _logger.debug('simplify %s', stmt.name)
      # LocalStmt and OutputStmt must remember the stencil object
      # for type propagation
      stmt.stencil = self
      stmt.expr = arithmetic.simplify(stmt.expr)
      stmt.let = arithmetic.simplify(stmt.let)

    self._cr_counter = 0
    cr.computation_reuse(self)
    if 'inline' in self.optimizations:
      inline.inline(self)
    inline.rebalance(self)

    for stmt in itertools.chain(self.local_stmts, self.output_stmts):
      stmt.propagate_type()

    # soda frontend successfully parsed
    _logger.debug('producer tensors: [%s]',
                  ', '.join(tensor.name for tensor in self.producer_tensors))
    _logger.debug('consumer tensors: [%s]',
                  ', '.join(tensor.name for tensor in self.consumer_tensors))

    # TODO: build Ref table and Var table
    # generate reuse buffers and get haoda nodes
    # pylint: disable=pointless-statement
    self.dataflow_super_source
    _logger.debug('dataflow: %s', self.dataflow_super_source)

    _logger.debug('module table: %s', dict(self.module_table))
    _logger.debug('module traits: %s', self.module_traits)

  def __str__(self) -> str:
    stmts = (self.input_stmts + self.param_stmts + self.local_stmts +
             self.output_stmts)
    return '''kernel: {0.app_name}
burst width: {0.burst_width}
iterate: {0.iterate}
unroll factor: {0.unroll_factor}
{stmts}
border: {0.border}
cluster: {0.cluster}'''.format(self, stmts='\n'.join(map(str, stmts)))

  @cached_property.cached_property
  def dataflow_super_source(self):
    return dataflow.create_dataflow_graph(self)

  @property
  def module_table(self):
    return self.dataflow_super_source.module_table

  @property
  def module_traits(self):
    return self.dataflow_super_source.module_traits

  def new_cr_var(self) -> str:
    while True:
      var = 'cr_var_%d' % (self._cr_counter)
      self._cr_counter += 1
      if var not in {
          stmt.name
          for stmt in self.input_stmts + self.local_stmts + self.output_stmts
      }:
        return var

  @cached_property.cached_property
  def stmt_table(
      self
  ) -> Dict[str, Union[soda.grammar.InputStmt, soda.grammar.LocalStmt,
                       soda.grammar.OutputStmt, soda.grammar.ParamStmt]]:
    return {
        stmt.name: stmt for stmt in self.input_stmts + self.local_stmts +
        self.output_stmts + self.param_stmts
    }

  @cached_property.cached_property
  def input_types(self):
    return tuple(tensor.haoda_type for tensor in self.input_stmts)

  @cached_property.cached_property
  def param_types(self):
    return tuple(tensor.haoda_type for tensor in self.param_stmts)

  @cached_property.cached_property
  def local_types(self):
    return tuple(tensor.haoda_type for tensor in self.local_stmts)

  @cached_property.cached_property
  def output_types(self):
    return tuple(tensor.haoda_type for tensor in self.output_stmts)

  @cached_property.cached_property
  def input_names(self):
    return tuple(stmt.name for stmt in self.input_stmts)

  @cached_property.cached_property
  def param_names(self):
    return tuple(stmt.name for stmt in self.param_stmts)

  @cached_property.cached_property
  def local_names(self):
    return tuple(stmt.name for stmt in self.local_stmts)

  @cached_property.cached_property
  def output_names(self):
    return tuple(stmt.name for stmt in self.output_stmts)

  @cached_property.cached_property
  def symbol_table(self) -> Dict[str, str]:
    """Constructs a dict mapping tensors' names to haoda types.

    Constructs a symbol table mapping names of the tensors to their haoda types.
    This table does not contain the local variables (ir.Let) in the LocalStmt
    and OutputStmt. For those, use stmt.symbol_table.

    Returns:
      symbol_table: A dict mapping tensors' names (str) to haoda types (str).

    Raises:
      util.InputError if names conflict.
    """
    symbol_table: Dict[str, str] = {}
    for name, haoda_type in zip(
        itertools.chain(self.input_names, self.local_names, self.output_names),
        itertools.chain(self.input_types, self.local_types, self.output_types)):
      if name in symbol_table:
        raise util.InputError('conflicting stmt name: %s' % name)
      symbol_table[name] = haoda_type
    return symbol_table

  @property
  def propagate_type(self):
    """Returns a callable that propagates type, optionally for a specific stmt.

    Returns:
      propagate_type: A callable that propagates type. It takes an optional
          argument of stmt, used for propagating types including the local
          variables (ir.Let).
    """

    def propagate_type(node, stmt=None):
      symbol_table = self.symbol_table
      if stmt is not None:
        symbol_table = stmt.symbol_table
      return arithmetic.base.propagate_type(node, symbol_table)

    return propagate_type

  @cached_property.cached_property
  def norm_refs(self):
    norm_refs = {}

    def get_norm_idx(stmt) -> Tuple[int, ...]:
      norm_idx = norm_refs.get(stmt.name)
      if norm_idx is None:
        loads = visitor.get_load_tuple(stmt.expr)
        for let in stmt.let:
          loads += visitor.get_load_tuple(let)

        def all_indices():
          for load in loads:
            if load.name in self.input_names:
              yield load.idx
            else:
              yield tuple(x + y for x, y in zip(
                  load.idx, get_norm_idx(self.stmt_table[load.name])))

        norm_idx = tuple(x - y for x, y in zip(
            min(all_indices(), key=lambda idx: tuple(reversed(tuple(idx)))),
            stmt.ref.idx))
        norm_refs[stmt.name] = norm_idx
        _logger.debug('%s has norm idx %s', stmt.name, norm_idx)
      return norm_idx

    for stmt in self.local_stmts + self.output_stmts:
      get_norm_idx(stmt)
    _logger.debug('norm refs %s', norm_refs)
    return norm_refs

  @cached_property.cached_property
  def tensors(self):
    """Constructs high-level DAG and creates the tensors.

    Returns:
      An collections.OrderedDict mapping a tensor's name to the tensor.
    """
    # TODO: check for name conflicts
    tensor_map = collections.OrderedDict()
    for stmt in self.input_stmts:
      tensor = soda.tensor.Tensor(stmt, self.tile_size)
      tensor_map[stmt.name] = tensor

    def name_in_iter(name, iteration):
      if name in self.input_names:
        if iteration > 0:
          return name + '_iter%d' % iteration
        return name
      if name in self.output_names:
        if iteration < self.iterate - 1:
          return (self.input_names[self.output_names.index(name)] + '_iter%d' %
                  (iteration + 1))
        return name
      if name in self.local_names:
        if iteration > 0:
          return name + '_iter%d' % iteration
        return name
      if name in self.param_names:
        return name
      raise util.InternalError('unknown name: %s' % name)

    for iteration in range(self.iterate):
      _logger.debug('iterate %s', iteration)
      _logger.debug('map: %s', self.symbol_table)

      def mutate_name_callback(obj, mutated):
        if isinstance(obj, ir.Ref):
          obj.haoda_type = self.symbol_table[obj.name]
          # pylint: disable=cell-var-from-loop
          obj.name = name_in_iter(obj.name, iteration)
        return obj

      tensors = []
      for stmt in itertools.chain(self.local_stmts, self.output_stmts):
        tensor = soda.tensor.Tensor(stmt.visit(mutate_name_callback),
                                    self.tile_size)
        tensor_map[tensor.name] = tensor
        tensors.append(tensor)

      for tensor in tensors:
        _logger.debug('%s', tensor)

      for tensor in tensors:
        tensor.propagate_type()
        loads = visitor.get_load_dict(tensor)
        for parent_name, ld_refs in loads.items():
          ld_refs = sorted(
              ld_refs,
              key=lambda ref: soda.util.serialize(ref.idx, self.tile_size))
          parent_tensor = tensor_map[parent_name]
          parent_tensor.children[tensor.name] = tensor
          tensor.parents[parent_name] = parent_tensor
          tensor.ld_refs[parent_name] = ld_refs

    # solve ILP for optimal reuse buffer
    lp_problem = pulp.LpProblem("optimal_reuse_buffer", pulp.LpMinimize)
    lp_vars = {self.input_names[0]: 0}  # set 1 and only 1 reference point
    lp_helper_vars = {}  # type: Dict[str, pulp.LpVariable]
    objectives = []
    constraints = []
    for tensor in tensor_map.values():
      lp_var = pulp.LpVariable('produced_offset_' + tensor.name, cat='Integer')
      lp_helper_var = pulp.LpVariable('consumed_offset_' + tensor.name,
                                      cat='Integer')
      lp_vars.setdefault(tensor.name, lp_var)
      lp_helper_vars[tensor.name] = lp_helper_var
      # tensor need to be kept for this long
      objectives.append(lp_helper_var - lp_vars[tensor.name])
      # tensor cannot be consumed until it is produced
      constraints.append(lp_helper_var >= lp_vars[tensor.name])
    lp_problem += sum(objectives)
    lp_problem.extend(constraints)
    for st_tensor in tensor_map.values():
      for ld_tensor_name, offsets in st_tensor.ld_offsets.items():
        oldest_access = min(offsets)
        newest_access = max(offsets)
        _logger.debug('%s @ %s accesses %s @ [%s, %s]', st_tensor.name,
                      st_tensor.st_offset, ld_tensor_name, oldest_access,
                      newest_access)
        # newest ld_tensor access must have been produced
        # when st_tensor is produced
        lp_problem += lp_vars[ld_tensor_name] <= lp_vars[st_tensor.name] + (
            st_tensor.st_offset - newest_access)
        # oldest ld_tensor access must have been not consumed
        # when st_tensor is produced
        lp_problem += lp_helper_vars[ld_tensor_name] >= lp_vars[
            st_tensor.name] + (st_tensor.st_offset - oldest_access)

    lp_status = lp_problem.solve()
    lp_status_str = pulp.LpStatus[lp_status]
    total_distance = int(pulp.value(lp_problem.objective))
    _logger.debug('ILP status: %s %s', lp_status_str, total_distance)
    _logger.info('total reuse distance: %d', total_distance)

    if lp_status != pulp.LpStatusOptimal:
      _logger.error('ILP error: %s\n%s', lp_status_str, lp_problem)
      raise util.InternalError('unexpected ILP status: %s' % lp_status_str)

    # some inputs may need to be delayed relative to others
    base = min(int(pulp.value(lp_vars[x])) for x in self.input_names)

    # set produce offsets
    for tensor in tensor_map.values():
      produce_offset = int(pulp.value(lp_vars[tensor.name])) - base
      consume_offset = int(pulp.value(lp_helper_vars[tensor.name])) - base
      tensor.produce_offset = produce_offset
      tensor.consume_offset = consume_offset
      tensor.max_access = 0  # pixels before current produce
      _logger.debug('%s should be produced @ %d and kept until %d', tensor.name,
                    produce_offset, consume_offset)

    # calculate overall acceses
    for ld_tensor in tensor_map.values():
      for st_tensor in ld_tensor.children.values():
        oldest_access = st_tensor.st_offset - min(
            st_tensor.ld_offsets[ld_tensor.name]
        ) + st_tensor.produce_offset - ld_tensor.produce_offset
        newest_access = st_tensor.st_offset - max(
            st_tensor.ld_offsets[ld_tensor.name]
        ) + st_tensor.produce_offset - ld_tensor.produce_offset
        _logger.debug(
            '  producing %s @ %s accesses [%s, %s] pixels before %s '
            'produced @ %s', st_tensor.name, st_tensor.produce_offset,
            newest_access, oldest_access, ld_tensor.name,
            ld_tensor.produce_offset)
        ld_tensor.max_access = max(ld_tensor.max_access, oldest_access)

    for tensor in tensor_map.values():
      _logger.debug('%s should be kept for %s pixels', tensor.name,
                    tensor.max_access)

    # high-level DAG construction finished
    for tensor in tensor_map.values():
      if tensor.name in self.input_names:
        _logger.debug('<input tensor>: %s', tensor)
      elif tensor.name in self.output_names:
        _logger.debug('<output tensor>: %s', tensor)
      else:
        _logger.debug('<local tensor>: %s', tensor)
    return tensor_map

  @cached_property.cached_property
  def chronological_tensors(self) -> List[soda.tensor.Tensor]:
    """Computes the offsets of tensors.

    Returns:
      A list of Tensor, in chronological order.
    """
    return list(
        map(
            self.tensors.get,
            toposort.toposort_flatten({
                tensor.name: set(tensor.parents)
                for tensor in self.tensors.values()
            })))

  @cached_property.cached_property
  def input_partition(self):
    pixel_width_i = sum(self.pixel_width_i)
    if (self.burst_width / pixel_width_i * self.dram_bank / 2 >
        self.unroll_factor / 2):
      return int(self.burst_width / pixel_width_i * self.dram_bank / 2)
    return int(self.unroll_factor / 2)

  @cached_property.cached_property
  def output_partition(self):
    pixel_width_o = sum(self.pixel_width_o)
    if (self.burst_width / pixel_width_o * self.dram_bank / 2 >
        self.unroll_factor / 2):
      return int(self.burst_width / pixel_width_o * self.dram_bank / 2)
    return int(self.unroll_factor / 2)

  @cached_property.cached_property
  def pixel_width_i(self):
    return [x.width_in_bits for x in self.input_stmts]

  @cached_property.cached_property
  def pixel_width_o(self):
    return [x.width_in_bits for x in self.output_stmts]

  @cached_property.cached_property
  def producer_tensors(self):
    return tuple(filter(soda.tensor.Tensor.is_producer, self.tensors.values()))

  @cached_property.cached_property
  def consumer_tensors(self):
    return tuple(filter(soda.tensor.Tensor.is_consumer, self.tensors.values()))

  @cached_property.cached_property
  def reuse_buffers(self):
    """Constructs the reuse buffers.

    Returns:
      A dict mapping a tensor's name to its reuse buffers.
    """
    unroll_factor = self.unroll_factor
    self._reuse_buffer_lengths = {}
    reuse_buffers = {}
    for tensor in self.producer_tensors:
      reuse_buffer = _get_reuse_buffer(self.tile_size, tensor, unroll_factor)
      reuse_buffer_length = {}
      reuse_buffers[tensor.name] = reuse_buffer
      self._reuse_buffer_lengths[tensor.name] = reuse_buffer_length
      first = [True] * unroll_factor
      for start, end in reuse_buffer[1:]:
        if first[start % unroll_factor]:
          first[start % unroll_factor] = False
          if start >= unroll_factor:
            reuse_buffer_length[end] = end // unroll_factor
            continue
        reuse_buffer_length[end] = (end - start) // unroll_factor
    return reuse_buffers

  @cached_property.cached_property
  def all_points(self):
    all_points = {}
    for tensor in self.producer_tensors:
      all_points[tensor.name] = _get_points(self.tile_size, tensor,
                                            self.unroll_factor)
    return all_points

  @cached_property.cached_property
  def next_fifo(self):
    """Constructs the next fifo offset mapping.

    Returns:
      A dict mapping a tensor's name and offset to the next offset.
    """
    next_fifo = {}
    for name, reuse_buffer in self.reuse_buffers.items():
      next_fifo[name] = {}
      for start, end in reuse_buffer[1:]:
        if start < end:
          next_fifo[name][start] = end
    _logger.debug('next_fifo: %s' % next_fifo)
    return next_fifo

  @cached_property.cached_property
  def reuse_buffer_lengths(self):
    """Constructs the reuse buffer lengths.

    Returns:
      A dict mapping a tensor's name to its reuse buffers' lengths.
    """
    # pylint: disable=pointless-statement
    self.reuse_buffers
    return self._reuse_buffer_lengths

  def get_replicated_next_fifo(self):
    if not hasattr(self, 'replicated_next_fifo'):
      self.replicated_next_fifo = {}
      for name, reuse_buffer in self.get_replicated_reuse_buffers().items():
        self.replicated_next_fifo[name] = {}
        for start, end in reuse_buffer[1:]:
          if start < end:
            self.replicated_next_fifo[name][start] = end
      _logger.debug('replicated_next_fifo: %s' % self.replicated_next_fifo)
    return self.replicated_next_fifo

  def get_replicated_reuse_buffer_length(self, name, offset):
    if not hasattr(self, 'replicated_reuse_buffer_lengths'):
      self.get_replicated_reuse_buffers()
    return self.replicated_reuse_buffer_lengths[name][offset]

  def get_replicated_reuse_buffers(self):
    if not hasattr(self, 'replicated_reuse_buffers'):
      replication_factor = self.replication_factor
      self.replicated_reuse_buffer_lengths = {}
      self.replicated_reuse_buffers = {}
      for tensor in self.producer_tensors:
        reuse_buffer = _get_replicated_reuse_buffer(self.tile_size, tensor,
                                                    replication_factor)
        self.replicated_reuse_buffers[tensor.name] = reuse_buffer
        self.replicated_reuse_buffer_lengths[tensor.name] = {}
        first = [True] * self.replication_factor
        for start, end in reuse_buffer[1:]:
          if first[start % replication_factor]:
            first[start % replication_factor] = False
            if start >= replication_factor:
              self.replicated_reuse_buffer_lengths[
                  tensor.name][end] = end // replication_factor
              continue
          self.replicated_reuse_buffer_lengths[
              tensor.name][end] = (end - start) // replication_factor
      _logger.debug('replicated_reuse_buffers: %s' %
                    self.replicated_reuse_buffers)
      _logger.debug('replicated_reuse_buffer_lengths: %s' %
                    self.replicated_reuse_buffer_lengths)
    return self.replicated_reuse_buffers

  def get_replicated_all_points(self):
    if not hasattr(self, 'replicated_all_points'):
      self.replicated_all_points = {}
      for tensor in self.producer_tensors:
        self.replicated_all_points[tensor.name] = _get_replicated_points(
            self.tile_size, tensor)
      _logger.debug('replicated_all_points: %s' % self.replicated_all_points)
    return self.replicated_all_points


def _get_reuse_chains(tile_size, tensor, unroll_factor):
  """Generates reuse chains for a Tensor.

  Generates reuse chains for a Tensor under the given tile size and unroll
  factor.

  Args:
    tile_size: An iterable representing the tile size in each dimension.
    tensor: A Tensor to which the reuse chains belongs.
    unroll_factor: An int representing the unroll factor.

  Returns:
    A list of tuples where each tuple represents a reuse chain and each
    element of the tuple represents the offset from the lastest input.
  """

  _logger.debug('get reuse chains of tensor %s', tensor.name)

  def unroll_offsets(child):
    unrolled_offsets = set()
    for unroll_idx in range(unroll_factor):
      for offset in child.ld_offsets[tensor.name]:
        unrolled_offsets.add(unroll_idx + child.st_offset - offset +
                             child.produce_offset - tensor.produce_offset)
    return unrolled_offsets

  A_dag = set()
  for child in tensor.children.values():
    A_dag |= unroll_offsets(child)
  _logger.debug('Aâ€  of tensor %s: %s', tensor.name, A_dag)

  chains = []
  for chain_idx in reversed(range(unroll_factor)):
    chains.append(
        tuple(
            sorted(offset for offset in A_dag
                   if offset % unroll_factor == chain_idx)))
  _logger.debug('reuse chains: %s', chains)

  for idx, chain in enumerate(chains):
    _logger.debug('reuse chain %d of tensor %s: %s', idx, tensor.name, chain)
  return chains


def _get_points(tile_size, tensor, unroll_factor):
  """Generates offset-to-point mapping for a Tensor.

  Generates a mapping which can be used to determine the accessed point index
  from the offset for a Tensor, under the given tile size and unroll factor.

  Args:
    tile_size: An iterable representing the tile size in each dimension.
    tensor: A Tensor to which the mapping belongs.
    unroll_factor: An int representing the unroll factor.

  Returns:
    A dict of name str to a dict of offset to a dict of unroll index to
    point index.
  """

  all_points = {}  # {name: {offset: {unroll_idx: point_idx}}}
  for child in tensor.children.values():
    all_points[child.name] = {}
    offsets = child.ld_offsets[tensor.name]
    for unroll_idx in range(unroll_factor):
      for idx, offset in enumerate(offsets):
        all_points[child.name].setdefault(
            unroll_idx + child.st_offset - offset + child.produce_offset -
            tensor.produce_offset, {})[unroll_factor - 1 - unroll_idx] = idx
  for child in tensor.children.values():
    for offset, points in all_points[child.name].items():
      for unroll_idx, point in points.items():
        _logger.debug(
            '%s <- %s @ offset=%d <=> %s @ unroll_idx=%d', child.name,
            tensor.name, offset,
            util.idx2str(
                list(child.ld_indices[tensor.name].values())[point].idx),
            unroll_idx)
  return all_points


def _get_reuse_buffer(tile_size, tensor, unroll_factor):
  """Generates reuse buffer for a Tensor.

  Generates a list representing the reuse buffer for a Tensor, under the given
  tile size and unroll factor.

  Args:
    tile_size: An iterable representing the tile size in each dimension.
    tensor: A Tensor to which the mapping belongs.
    unroll_factor: An int representing the unroll factor.

  Returns:
    A list whose first element is an int representing the length of the
    reuse buffer (capacity of data element), followed by unroll_factor
    number of (start, end) tuples, where start and end are the offsets from
    the lastest input of each piece of the reuse buffer.
  """

  reuse_buffer = [None]  # [length, (start, end), (start, end), ...]
  offsets = []
  for chain_id, chain in enumerate(
      _get_reuse_chains(tile_size, tensor, unroll_factor)):
    reuse_buffer.append((unroll_factor - 1 - chain_id, chain[0]))
    _logger.debug('chain id: %s, chain[0]: %s', chain_id, chain[0])
    offsets.append(chain[0])
    for j in range(len(chain) - 1):
      reuse_buffer.append((chain[j], chain[j + 1]))
      offsets.append(chain[j + 1])
  reuse_buffer[0] = max(offsets) + 1
  _logger.debug('reuse chains of tensor %s: %s' % (tensor.name, reuse_buffer))
  return reuse_buffer


def _get_replicated_reuse_chains(tile_size, tensor, replication_factor):
  _logger.debug('\033[1mget replicated reuse chains of tensor %s\033[0m' %
                tensor.name)
  A_dag = set()
  for stage in tensor.children:
    offsets = soda.util.serialize_iter(stage.window[tensor.name], tile_size)
    A_dag |= {
        max(offsets) - offset + stage.delay[tensor.name] for offset in offsets
    }
  _logger.debug('Aâ€  of tensor %s: %s' % (tensor.name, A_dag))
  chains = sum(
      reversed([
          tuple(
              [tuple(sorted(x
                            for x in A_dag
                            if x % replication_factor == i))])
          for i in range(replication_factor)
      ]), ())
  for idx, chain in enumerate(chains):
    _logger.debug('reuse chain %d of tensor %s: %s' % (idx, tensor.name, chain))
  return chains


def _get_replicated_points(tile_size, tensor):
  all_points = {}  # {name:{offset:point_index}}
  for stage in tensor.children:
    all_points[stage.name] = {}
    offsets = soda.util.serialize_iter(stage.window[tensor.name], tile_size)
    max_offset = max(offsets)
    for idx, offset in enumerate(offsets):
      all_points[stage.name][max_offset - offset +
                             stage.delay[tensor.name]] = idx
  for stage in tensor.children:
    for offset, points in all_points[stage.name].items():
      _logger.debug('%s <- %s @ offset=%d <=> (%s)' %
                    (stage.name, tensor.name, offset, ', '.join(
                        map(str, stage.window[tensor.name][points]))))
  return all_points


def _get_replicated_reuse_buffer(tile_size, tensor, replication_factor):
  reuse_buffer = [None]  # [length, (start, end), (start, end), ...]
  offsets = []
  for chain in _get_replicated_reuse_chains(tile_size, tensor,
                                            replication_factor):
    if chain:
      reuse_buffer.append((chain[0], chain[0]))
      offsets.append(chain[0])
      for j in range(len(chain) - 1):
        reuse_buffer.append((chain[j], chain[j + 1]))
        offsets.append(chain[j + 1])
  reuse_buffer[0] = max(offsets) + 1
  _logger.debug('reuse chains of tensor %s: %s' % (tensor.name, reuse_buffer))
  return reuse_buffer


def get_indices_id(indices):
  return '_'.join(str(idx).replace('-', 'm') for idx in indices)


def get_stencil_distance(stencil_window, tile_size):
  return (
      max(soda.util.serialize_iter(stencil_window, tile_size)) +
      soda.util.serialize(get_stencil_window_offset(stencil_window), tile_size))


def get_stencil_dim(points):
  dimension = len(next(iter(points)))
  return [
      max_index - min_index + 1 for max_index, min_index in
      zip([max([point[dim] for point in points]) for dim in range(dimension)],
          [min([point[dim] for point in points]) for dim in range(dimension)])
  ]

_overall_stencil_window_cache = {} \
    # type: Dict[Tuple[int, int], Tuple[Tuple[int, ...], ...]]


def get_overall_stencil_window(input_tensor, output_tensor):
  if isinstance(input_tensor, collections.Iterable):
    all_points = tuple(
        sorted(
            set.union(*(set(get_overall_stencil_window(_, output_tensor))
                        for _ in input_tensor))))
    _logger.debug('overall stencil window of %s (%s) <- {%s} is %s (%d points)',
                  output_tensor.name,
                  ', '.join(['0'] * len(output_tensor.st_idx)),
                  ', '.join(_.name for _ in input_tensor), all_points,
                  len(all_points))
    return all_points
  # normalize store index to 0
  idx = (id(input_tensor), id(output_tensor))
  if idx in _overall_stencil_window_cache:
    return _overall_stencil_window_cache[idx]
  _logger.debug('get overall stencil window of %s <- %s', output_tensor.name,
                input_tensor.name)
  all_points = set()
  for name, points in output_tensor.ld_indices.items():
    _logger.debug('%s@%s <- %s', output_tensor.name,
                  util.idx2str(output_tensor.st_idx),
                  util.idx2str(points.values()))
    if name != input_tensor.name:
      recursive_points = get_overall_stencil_window(input_tensor,
                                                    output_tensor.parents[name])
      _logger.debug('recursive points: %s', util.idx2str(recursive_points))
      all_points |= set.union(*[{
          tuple(map(lambda a, b, c: a + b - c, _, point, output_tensor.st_idx))
          for _ in recursive_points
      }
                                for point in points])
    else:
      all_points |= {
          tuple(map(operator.sub, point, output_tensor.st_idx))
          for point in points
      }
  all_points = tuple(sorted(all_points))
  _logger.debug('overall stencil window of %s (%s) <- %s is %s (%d points)',
                output_tensor.name,
                ', '.join(['0'] * len(output_tensor.st_idx)), input_tensor.name,
                all_points, len(all_points))
  _overall_stencil_window_cache[idx] = all_points
  return all_points


def get_stencil_window_offset(stencil_window):
  # only works if window is normalized to store at 0
  return tuple(-min(p[d]
                    for p in stencil_window)
               for d in range(len(next(iter(stencil_window)))))



import collections
import concurrent.futures
import io
import json
import logging
import os
import shutil
import sys
import tarfile
import tempfile
from typing import (IO, Dict, Iterable, List, Optional, Sequence, Set, TextIO,
                    Tuple)

from haoda import util
from haoda.backend import xilinx as backend
from haoda.report.xilinx import hls as hls_report
from soda import core, dataflow
from soda.codegen.xilinx import hls_kernel

_logger = logging.getLogger().getChild(__name__)


def print_code(
    stencil: core.Stencil,
    xo_file: IO[bytes],
    device_info: Dict[str, str],
    jobs: Optional[int] = os.cpu_count(),
    rpt_file: Optional[str] = None,
    interface: str = 'm_axi',
) -> None:
  """Generate hardware object file for the given Stencil.

  Working `vivado` and `vivado_hls` is required in the PATH.

  Args:
    stencil: Stencil object to generate from.
    xo_file: file object to write to.
    device_info: dict of 'part_num' and 'clock_period'.
    jobs: maximum number of jobs running in parallel.
    rpt_file: path of the generated report; None disables report generation.
    interface: interface type, supported values are 'm_axi' and 'axis'.
  """

  iface_names = []  # for axis
  m_axi_names = []  # for m_axi
  inputs = []
  outputs = []

  for stmt in stencil.output_stmts:
    for bank in stmt.dram:
      port_name = util.get_port_name(stmt.name, bank)
      bundle_name = util.get_bundle_name(stmt.name, bank)
      iface_names.append(port_name)
      m_axi_names.append(bundle_name)
      outputs.append((port_name, bundle_name, stencil.burst_width,
                      util.get_port_buf_name(stmt.name, bank)))
  for stmt in stencil.input_stmts:
    for bank in stmt.dram:
      port_name = util.get_port_name(stmt.name, bank)
      bundle_name = util.get_bundle_name(stmt.name, bank)
      iface_names.append(port_name)
      m_axi_names.append(bundle_name)
      inputs.append((port_name, bundle_name, stencil.burst_width,
                     util.get_port_buf_name(stmt.name, bank)))

  top_name = stencil.app_name + '_kernel'
  with tempfile.TemporaryDirectory(prefix='sodac-xrtl-') as tmpdir:
    kernel_xml = os.path.join(tmpdir, 'kernel.xml')
    with open(kernel_xml, 'w') as kernel_xml_obj:
      print_kernel_xml(top_name, inputs, outputs, kernel_xml_obj, interface)

    kernel_file = os.path.join(tmpdir, 'kernel.cpp')
    with open(kernel_file, 'w') as kernel_fileobj:
      hls_kernel.print_code(stencil, kernel_fileobj)

    args = []
    for module_trait_id, module_trait in enumerate(stencil.module_traits):
      sio = io.StringIO()
      hls_kernel.print_module_definition(util.CppPrinter(sio),
                                         module_trait,
                                         module_trait_id,
                                         burst_width=stencil.burst_width)
      args.append((len(sio.getvalue()), synthesis_module, tmpdir, [kernel_file],
                   util.get_func_name(module_trait_id), device_info))

    if interface == 'm_axi':
      sio = io.StringIO()
      print_dataflow_hls_interface(util.CppPrinter(sio), top_name, inputs,
                                   outputs)
      dataflow_kernel = os.path.join(tmpdir, 'dataflow_kernel.cpp')
      with open(dataflow_kernel, 'w') as dataflow_kernel_obj:
        dataflow_kernel_obj.write(sio.getvalue())
      args.append((len(sio.getvalue()), synthesis_module, tmpdir,
                   [dataflow_kernel], top_name, device_info))
    args.sort(key=lambda x: x[0], reverse=True)

    super_source = stencil.dataflow_super_source
    job_server = util.release_job_slot()
    with concurrent.futures.ThreadPoolExecutor(max_workers=jobs) as executor:
      threads = [executor.submit(*x[1:]) for x in args]
      for future in concurrent.futures.as_completed(threads):
        returncode, stdout, stderr = future.result()
        log_func = _logger.error if returncode != 0 else _logger.debug
        if stdout:
          log_func(stdout.decode())
        if stderr:
          log_func(stderr.decode())
        if returncode != 0:
          util.pause_for_debugging()
          sys.exit(returncode)
    util.acquire_job_slot(job_server)

    # generate HLS report
    depths: Dict[int, int] = {}
    hls_resources = hls_report.HlsResources()
    if interface == 'm_axi':
      hls_resources = hls_report.resources(
          os.path.join(tmpdir, 'report', top_name + '_csynth.xml'))
      hls_resources -= hls_report.resources(
          os.path.join(tmpdir, 'report', 'Dataflow_csynth.xml'))

    _logger.info(hls_resources)
    for module_id, nodes in enumerate(super_source.module_trait_table.values()):
      module_name = util.get_func_name(module_id)
      report_file = os.path.join(tmpdir, 'report', module_name + '_csynth.xml')
      hls_resource = hls_report.resources(report_file)
      use_count = len(nodes)
      try:
        perf = hls_report.performance(report_file)
        _logger.info('%s, usage: %5d times, II: %3d, Depth: %3d', hls_resource,
                     use_count, perf.ii, perf.depth)
        depths[module_id] = perf.depth
      except hls_report.BadReport as e:
        _logger.warn('%s in %s report (%s)', e, module_name, report_file)
        _logger.info('%s, usage: %5d times', hls_resource, use_count)
        raise e
      hls_resources += hls_resource * use_count
    _logger.info('total usage:')
    _logger.info(hls_resources)
    if rpt_file:
      rpt_json = collections.OrderedDict([('name', top_name)] +
                                         list(hls_resources))
      with open(rpt_file, mode='w') as rpt_fileobj:
        json.dump(rpt_json, rpt_fileobj, indent=2)

    # update the module pipeline depths
    stencil.dataflow_super_source.update_module_depths(depths)

    hdl_dir = os.path.join(tmpdir, 'hdl')
    module_name = 'Dataflow'
    if interface == 'axis':
      module_name = top_name
    with open(os.path.join(hdl_dir, f'{module_name}.v'), mode='w') as fileobj:
      print_top_module(
          backend.VerilogPrinter(fileobj),
          stencil.dataflow_super_source,
          inputs,
          outputs,
          module_name,
          interface,
      )

    util.pause_for_debugging()

    xo_filename = os.path.join(tmpdir, stencil.app_name + '.xo')
    kwargs = {}
    if interface == 'm_axi':
      kwargs['m_axi_names'] = m_axi_names
    elif interface == 'axis':
      kwargs['iface_names'] = iface_names
    with backend.PackageXo(
        xo_filename,
        top_name,
        kernel_xml,
        hdl_dir,
        **kwargs,
    ) as proc:
      stdout, stderr = proc.communicate()
    log_func = _logger.error if proc.returncode != 0 else _logger.debug
    log_func(stdout.decode())
    log_func(stderr.decode())
    with open(xo_filename, mode='rb') as xo_fileobj:
      shutil.copyfileobj(xo_fileobj, xo_file)


def synthesis_module(tmpdir, kernel_files, module_name, device_info):
  """Synthesis a module in kernel files.

  Returns:
    (returncode, stdout, stderr) results of the subprocess.
  """
  job_server = util.acquire_job_slot()
  with tempfile.TemporaryFile(mode='w+b') as tarfileobj:
    with backend.RunHls(tarfileobj,
                        kernel_files,
                        module_name,
                        device_info['clock_period'],
                        device_info['part_num'],
                        reset_low=False) as proc:
      stdout, stderr = proc.communicate()
    if proc.returncode == 0:
      tarfileobj.seek(0)
      with tarfile.open(mode='r', fileobj=tarfileobj) as tar:
        tar.extractall(
            tmpdir,
            (f for f in tar.getmembers()
             if f.name.startswith('hdl') or f.name.startswith('report')))
  util.release_job_slot(job_server)
  return proc.returncode, stdout, stderr


FIFO_PORT_SUFFIXES = dict(data_in='_din',
                          not_full='_full_n',
                          write_enable='_write',
                          data_out='_dout',
                          not_empty='_empty_n',
                          read_enable='_read',
                          not_block='_blk_n')

AXIS_PORT_SUFFIXES = collections.OrderedDict(
    data='_TDATA',  # producer -> consumer
    keep='_TKEEP',  # producer -> consumer
    strb='_TSTRB',  # producer -> consumer
    last='_TLAST',  # producer -> consumer
    valid='_TVALID',  # producer -> consumer
    ready='_TREADY',  # producer <- consumer
)


def print_kernel_xml(
    name: str,
    inputs: Iterable[Tuple[str, str, int, str]],
    outputs: Iterable[Tuple[str, str, int, str]],
    kernel_xml: TextIO,
    interface: str = 'm_axi',
) -> None:
  """Generate kernel.xml file.

  Args:
    name: Name of the kernel.
    inputs: Iterable of (port_name, bundle_name, width, _) of input ports.
    outputs: Iterable of (port_name, bundle_name, width, _) of output ports.
    kernel_xml: File object to write to.
    interface: Interface type, supported values are 'm_axi' and 'axis'.
  """
  args: List[backend.Arg] = []
  if interface == 'm_axi':
    for ports in outputs, inputs:
      for port_name, bundle_name, width, _ in ports:
        args.append(
            backend.Arg(
                cat=backend.Cat.MMAP,
                name=port_name,
                port=bundle_name,
                ctype=f'ap_uint<{width}>*',
                width=width,
            ))
    args.append(
        backend.Arg(
            cat=backend.Cat.SCALAR,
            name='coalesced_data_num',
            port='',
            ctype='uint64_t',
            width=64,
        ))
  elif interface == 'axis':
    for cat, ports in ((backend.Cat.ISTREAM, inputs), (backend.Cat.OSTREAM,
                                                       outputs)):
      for port_name, _, width, _ in ports:
        ctype = f'stream<ap_axiu<{width}, 0, 0, 0>>&'
        args.append(
            backend.Arg(
                cat=cat,
                name=port_name,
                port='',
                ctype=ctype,
                width=width,
            ))
  else:
    raise util.InternalError(f'unexpected interface `{interface}`')

  backend.print_kernel_xml(name=name, args=args, kernel_xml=kernel_xml)


def print_top_module(
    printer: backend.VerilogPrinter,
    super_source: dataflow.SuperSourceNode,
    inputs: Sequence[Tuple[str, str, int, str]],
    outputs: Sequence[Tuple[str, str, int, str]],
    module_name: str = 'Dataflow',
    interface: str = 'm_axi',
) -> None:
  """Generate kernel.xml file.

  Args:
    printer: printer to print to
    super_source: SuperSourceNode carrying the IR tree.
    inputs: sequence of (port_name, bundle_name, width, _) of input ports
    outputs: sequence of (port_name, bundle_name, width, _) of output ports
    module_name: name of the module
    interface: interface type, supported values are 'm_axi' and 'axis'
  """
  printer.printlns('`timescale 1 ns / 1 ps', '`default_nettype none')

  ports = *inputs, *outputs

  # unpack suffixes
  data_in = FIFO_PORT_SUFFIXES['data_in']
  not_full = FIFO_PORT_SUFFIXES['not_full']
  write_enable = FIFO_PORT_SUFFIXES['write_enable']
  data_out = FIFO_PORT_SUFFIXES['data_out']
  not_empty = FIFO_PORT_SUFFIXES['not_empty']
  read_enable = FIFO_PORT_SUFFIXES['read_enable']
  not_block = FIFO_PORT_SUFFIXES['not_block']
  data = AXIS_PORT_SUFFIXES['data']
  keep = AXIS_PORT_SUFFIXES['keep']
  strb = AXIS_PORT_SUFFIXES['strb']
  last = AXIS_PORT_SUFFIXES['last']
  valid = AXIS_PORT_SUFFIXES['valid']
  ready = AXIS_PORT_SUFFIXES['ready']

  # prepare arguments
  input_args = ['ap_clk']
  output_args: List[str] = []
  if interface == 'm_axi':
    input_args += 'ap_rst', 'ap_start', 'ap_continue'
    output_args += 'ap_done', 'ap_idle', 'ap_ready'
  elif interface == 'axis':
    input_args.append('ap_rst_n')

  args = list(input_args + output_args)
  if interface == 'm_axi':
    for port_name, _, _, _ in outputs:
      args.append(f'{port_name}_V_V{data_in}')
      args.append(f'{port_name}_V_V{not_full}')
      args.append(f'{port_name}_V_V{write_enable}')
    for port_name, _, _, _ in inputs:
      args.append(f'{port_name}_V_V{data_out}')
      args.append(f'{port_name}_V_V{not_empty}')
      args.append(f'{port_name}_V_V{read_enable}')
  elif interface == 'axis':
    for port_name, _, _, _ in ports:
      args.extend(port_name + suffix for suffix in AXIS_PORT_SUFFIXES.values())

  # print module interface
  printer.module(module_name, args)
  printer.println()

  # print signals for modules
  printer.printlns(
      *(f'input  wire {arg};' for arg in input_args),
      *(f'output wire {arg};' for arg in output_args),
  )
  for port_name, _, width, _ in outputs:
    if interface == 'm_axi':
      printer.printlns(
          f'output wire [{width - 1}:0] {port_name}_V_V{data_in};',
          f'input  wire {port_name}_V_V{not_full};',
          f'output wire {port_name}_V_V{write_enable};',
      )
    elif interface == 'axis':
      printer.printlns(
          f'output wire [{width - 1}:0] {port_name}{data};',
          f'output wire [{width // 8 - 1}:0] {port_name}{keep};',
          f'output wire [{width // 8 - 1}:0] {port_name}{strb};',
          f'output wire {port_name}{last};',
          f'output wire {port_name}{valid};',
          f'input  wire {port_name}{ready};',
          f'wire [{width - 1}:0] {port_name}_V_V{data_in};',
          f'wire {port_name}_V_V{not_full};',
          f'wire {port_name}_V_V{write_enable};',
      )
  for port_name, _, width, _ in inputs:
    if interface == 'm_axi':
      printer.printlns(
          f'input  wire [{width - 1}:0] {port_name}_V_V{data_out};',
          f'input  wire {port_name}_V_V{not_empty};',
          f'output wire {port_name}_V_V{read_enable};',
      )
    elif interface == 'axis':
      printer.printlns(
          f'input  wire [{width - 1}:0] {port_name}{data};',
          f'input  wire [{width // 8 - 1}:0] {port_name}{keep};',
          f'input  wire [{width // 8 - 1}:0] {port_name}{strb};',
          f'input  wire {port_name}{last};',
          f'input  wire {port_name}{valid};',
          f'output wire {port_name}{ready};',
          f'wire [{width - 1}:0] {port_name}_V_V{data_out};',
          f'wire {port_name}_V_V{not_empty};',
          f'wire {port_name}_V_V{read_enable};',
      )
  printer.println()

  # not used
  printer.printlns(
      "reg ap_done = 1'b0;",
      "reg ap_idle = 1'b1;",
      "reg ap_ready = 1'b0;",
  )
  if interface == 'axis':
    printer.println("wire ap_start = 1'b1;")

  # print signals for FIFOs
  if interface == 'm_axi':
    for port_name, _, width, _ in outputs:
      printer.printlns(
          f'reg [{width - 1}:0] {port_name}{data_in};',
          f'wire {port_name}_V_V{write_enable};',
      )
    for port_name, _, _, _ in inputs:
      printer.println(f'wire {port_name}_V_V{read_enable};')
  printer.println()

  # register reset signal
  ap_rst_reg_level = 8
  rst = 'ap_rst'
  if interface == 'axis':
    rst = '~ap_rst_n'
  printer.printlns(
      f'wire ap_rst_reg_0 = {rst};',
      *(f'(* shreg_extract = "no", max_fanout = {8 ** i} *) reg ap_rst_reg_{i};'
        for i in range(1, ap_rst_reg_level)),
      f'(* shreg_extract = "no" *) reg ap_rst_reg_{ap_rst_reg_level};',
      f'wire ap_rst_reg = ap_rst_reg_{ap_rst_reg_level};',
  )
  if ap_rst_reg_level > 0:
    with printer.always('posedge ap_clk'):
      printer.printlns(f'ap_rst_reg_{i + 1} <= ap_rst_reg_{i};'
                       for i in range(ap_rst_reg_level))

  with printer.always('posedge ap_clk'):
    with printer.if_('ap_rst_reg'):
      printer.printlns(
          "ap_done <= 1'b0;",
          "ap_idle <= 1'b1;",
          "ap_ready <= 1'b0;",
      )
      printer.else_()
      printer.println('ap_idle <= ~ap_start;')
  printer.println()

  if interface == 'm_axi':
    # used by cosim for deadlock detection
    printer.printlns(
        f'reg {port_name}_V_V{not_block};' for port_name, _, _, _ in ports)
    with printer.always('*'):
      printer.printlns(
          *(f'{port_name}_V_V{not_block} = {port_name}_V_V{not_full};'
            for port_name, _, _, _ in outputs),
          *(f'{port_name}_V_V{not_block} = {port_name}_V_V{not_empty};'
            for port_name, _, _, _ in inputs),
      )
  printer.println()

  fifos: Set[Tuple[int, int]] = set()  # used for printing FIFO modules

  if interface == 'axis':
    for port_name, _, width, _ in outputs:
      width = width // 8
      ones = '1' * width
      printer.printlns(
          f"assign {port_name}{keep} = {width}'b{ones};",
          f"assign {port_name}{strb} = {width}'b{ones};",
          f"assign {port_name}{last} = 1'b0;",
      )

    for port_name, _, width, _ in inputs:
      printer.module_instance(
          'fifo_w{width}_d{depth}_A'.format(width=width, depth=2),
          port_name + '_fifo',
          args={
              'clk': 'ap_clk',
              'reset': 'ap_rst_reg',
              'if_read_ce': "1'b1",
              'if_write_ce': "1'b1",
              f'if{data_in}': f'{port_name}{data}',
              f'if{not_full}': f'{port_name}{ready}',
              f'if{write_enable}': f'{port_name}{valid}',
              f'if{data_out}': f'{port_name}_V_V{data_out}',
              f'if{not_empty}': f'{port_name}_V_V{not_empty}',
              f'if{read_enable}': f'{port_name}_V_V{read_enable}',
          },
      )
      fifos.add((width, 2))
      printer.println()

    for port_name, _, width, _ in outputs:
      printer.module_instance(
          f'fifo_w{width}_d2_A',
          port_name + '_fifo',
          args={
              'clk': 'ap_clk',
              'reset': 'ap_rst_reg',
              'if_read_ce': "1'b1",
              'if_write_ce': "1'b1",
              f'if{data_in}': f'{port_name}_V_V{data_in}',
              f'if{not_full}': f'{port_name}_V_V{not_full}',
              f'if{write_enable}': f'{port_name}_V_V{write_enable}',
              f'if{data_out}': f'{port_name}{data}',
              f'if{not_empty}': f'{port_name}{valid}',
              f'if{read_enable}': f'{port_name}{ready}',
          },
      )
      fifos.add((width, 2))
      printer.println()

  # print FIFO instances
  for module in super_source.tpo_valid_node_gen():
    for fifo in module.fifos:
      name = fifo.c_expr
      msb = fifo.width_in_bits - 1
      # fifo.depth is the "extra" capacity of a FIFO; the base depth is 3, 1 for
      # registering the input, 1 for keeping II=1 when FIFO is (almost) full, 1
      # for keeping II=1 when FIFO is relaxed from back pressure (necessary
      # because the optimal FIFO depths may require back pressure)
      depth = fifo.depth + 3
      printer.printlns(
          f'wire [{msb}:0] {name}{data_in};',
          f'wire {name}{not_full};',
          f'wire {name}{write_enable};',
          f'wire [{msb}:0] {name}{data_out};',
          f'wire {name}{not_empty};',
          f'wire {name}{read_enable};',
          '',
      )
      printer.module_instance(
          f'fifo_w{fifo.width_in_bits}_d{depth}_A',
          name,
          args={
              'clk': 'ap_clk',
              'reset': 'ap_rst_reg',
              'if_read_ce': "1'b1",
              'if_write_ce': "1'b1",
              f'if{data_in}': f'{name}{data_in}',
              f'if{not_full}': f'{name}{not_full}',
              f'if{write_enable}': f'{name}{write_enable}',
              f'if{data_out}': f'{name}{data_out}',
              f'if{not_empty}': f'{name}{not_empty}',
              f'if{read_enable}': f'{name}{read_enable}',
          },
      )
      fifos.add((fifo.width_in_bits, depth))
      printer.println()

  # print module instances
  for module in super_source.tpo_valid_node_gen():
    module_trait, module_trait_id = super_source.module_table[module]
    arg_dict = {
        'ap_clk': 'ap_clk',
        'ap_rst': 'ap_rst_reg',
        'ap_start': "1'b1",
    }
    for dram_ref, bank in module.dram_writes:
      port = dram_ref.dram_fifo_name(bank)
      fifo = util.get_port_name(dram_ref.var, bank)
      arg_dict.update({
          f'{port}_V{data_in}': f'{fifo}_V_V{data_in}',
          f'{port}_V{not_full}': f'{fifo}_V_V{not_full}',
          f'{port}_V{write_enable}': f'{fifo}_V_V{write_enable}',
      })
    for port, fifo in zip(module_trait.output_fifos, module.output_fifos):
      arg_dict.update({
          f'{port}_V{data_in}': f'{fifo}{data_in}',
          f'{port}_V{not_full}': f'{fifo}{not_full}',
          f'{port}_V{write_enable}': f'{fifo}{write_enable}',
      })
    for port, fifo in zip(module_trait.input_fifos, module.input_fifos):
      arg_dict.update({
          f'{port}_V{data_out}': f"{{1'b1, {fifo}{data_out}}}",
          f'{port}_V{not_empty}': f'{fifo}{not_empty}',
          f'{port}_V{read_enable}': f'{fifo}{read_enable}',
      })
    for dram_ref, bank in module.dram_reads:
      port = dram_ref.dram_fifo_name(bank)
      fifo = util.get_port_name(dram_ref.var, bank)
      arg_dict.update({
          f'{port}_V{data_out}': f"{{1'b1, {fifo}_V_V{data_out}}}",
          f'{port}_V{not_empty}': f'{fifo}_V_V{not_empty}',
          f'{port}_V{read_enable}': f'{fifo}_V_V{read_enable}',
      })
    printer.module_instance(util.get_func_name(module_trait_id), module.name,
                            arg_dict)
    printer.println()
  printer.endmodule()
  printer.println('`default_nettype wire')

  # print FIFO modules
  for fifo in fifos:
    printer.fifo_module(*fifo)


def print_dataflow_hls_interface(
    printer: backend.util.CppPrinter,
    top_name: str,
    inputs: Sequence[Tuple[str, str, int, str]],
    outputs: Sequence[Tuple[str, str, int, str]],
) -> None:
  ports = *outputs, *inputs

  printer.printlns(
      '#include <cstddef>',
      '#include <cstdint>',
      '#include <ap_int.h>',
      '#include <hls_stream.h>',
  )

  printer.println('template<typename T>')
  printer.print_func('void BurstRead',
                     ['hls::stream<T>& to', 'T* from', 'uint64_t data_num'],
                     align=0)
  printer.do_scope()
  printer.println('burst_read:', 0)
  with printer.for_('uint64_t epoch = 0', 'epoch < data_num', '++epoch'):
    printer.println('#pragma HLS pipeline II=1', 0)
    printer.println('to.write(from[epoch]);')
  printer.un_scope()

  printer.println('template<typename T>')
  printer.print_func('void BurstWrite',
                     ['T* to', 'hls::stream<T>& from', 'uint64_t data_num'],
                     align=0)
  printer.do_scope()
  printer.println('burst_write:', 0)
  with printer.for_('uint64_t epoch = 0', 'epoch < data_num', '++epoch'):
    printer.println('#pragma HLS pipeline II=1', 0)
    printer.println('to[epoch] = from.read();')
  printer.un_scope()

  params = [
      f'hls::stream<ap_uint<{width}>>& {name}' for name, _, width, _ in ports
  ]
  printer.print_func('void Dataflow', params, align=0)
  printer.do_scope()
  printer.printlns(
      *(f'volatile ap_uint<{width}> {name}_read = {name}.read();'
        for name, _, width, _ in inputs),
      *(f'{name}.write(ap_uint<{width}>());' for name, _, width, _ in outputs),
  )
  printer.un_scope()

  params = [f'ap_uint<{width}>* {name}' for name, _, width, _ in ports]
  params.append('uint64_t coalesced_data_num')
  printer.print_func('void %s' % top_name, params, align=0)
  printer.do_scope()
  printer.println('#pragma HLS dataflow', 0)
  for port_name, bundle_name, _, _ in ports:
    printer.println(
        f'#pragma HLS interface m_axi port={port_name} '
        f'offset=slave bundle={bundle_name}', 0)
  for port_name, _, _, _ in ports:
    printer.println(
        f'#pragma HLS interface s_axilite port={port_name} '
        'bundle=control', 0)
  printer.println(
      '#pragma HLS interface s_axilite port=coalesced_data_num '
      'bundle=control', 0)
  printer.println(
      '#pragma HLS interface s_axilite port=return '
      'bundle=control', 0)
  printer.println()
  for _, _, width, name in ports:
    printer.println(f'hls::stream<ap_uint<{width}>> {name}("{name}");')
    printer.println(f'#pragma HLS stream variable={name} depth=32', 0)
  for port_name, _, _, buf_name in inputs:
    printer.print_func(
        'BurstRead',
        [buf_name, port_name, 'coalesced_data_num'],
        suffix=';',
        align=0,
    )
  printer.print_func(
      'Dataflow',
      [name for _, _, _, name in ports],
      suffix=';',
      align=0,
  )
  for port_name, _, _, buf_name in outputs:
    printer.print_func(
        'BurstWrite',
        [port_name, buf_name, 'coalesced_data_num'],
        suffix=';',
        align=0,
    )
  printer.un_scope()



import argparse
import os
import shutil
import sys
import tempfile

from haoda.backend import xilinx as backend
from soda import core
from soda.codegen.xilinx import header
from soda.codegen.xilinx import hls_kernel as kernel
from soda.codegen.xilinx import host, rtl_kernel

def add_arguments(parser):
  parser.add_argument(
      '--xocl', type=str, dest='output_dir', metavar='dir', nargs='?', const='',
      help='directory to generate kernel and host code; default names are'
      'used; default to the current working directory; may be overridden by '
      '--xocl-header, --xocl-host, or --xocl-kernel')
  parser.add_argument(
      '--xocl-header', type=str, dest='header_file', metavar='file',
      help='host C++ header code; overrides --xocl')
  parser.add_argument(
      '--xocl-host', type=str, dest='host_file', metavar='file',
      help='host C++ source code for the Xilinx OpenCL flow; overrides --xocl')
  parser.add_argument(
      '--xocl-kernel', type=str, dest='kernel_file', metavar='file',
      help='Vivado HLS C++ kernel code for the Xilinx OpenCL flow; overrides '
      '--xocl')
  parser.add_argument(
      '--xocl-platform', type=str, dest='xocl_platform', metavar='dir',
      help='SDAccel platform directory of the Xilinx OpenCL flow')
  parser.add_argument(
      '--xocl-part-num', type=str, dest='xocl_part_num', metavar='string',
      help='part number used for Vivado HLS')
  parser.add_argument(
      '--xocl-clock-period', type=str, dest='xocl_clock_period', metavar='ns',
      help='target clock period in nanoseconds, used for Vivado HLS')
  parser.add_argument('--xocl-hw-xo', type=str, dest='xo_file', metavar='file',
                      help='hardware object file for the Xilinx OpenCL flow')
  parser.add_argument(
      '--xocl-hw-xo-rpt', type=str, dest='xo_rpt', metavar='file',
      help='Vivado HLS report for the Xilinx OpenCL hardware object')
  parser.add_argument(
    '--xocl-interface', type=str, dest='interface', metavar='(m_axi|axis)',
    default='m_axi', help='interface type of the Xilinx OpenCL hardware object')

def print_code(
    stencil: core.Stencil,
    args: argparse.Namespace,
    parser: argparse.ArgumentParser,
) -> None:
  if args.kernel_file is not None:
    with tempfile.TemporaryFile(mode='w+') as tmp:
      kernel.print_code(stencil, tmp)
      tmp.seek(0)
      if args.kernel_file == '-':
        shutil.copyfileobj(tmp, sys.stdout)
      else:
        with open(args.kernel_file, 'w') as kernel_file:
          shutil.copyfileobj(tmp, kernel_file)

  if args.host_file is not None:
    with tempfile.TemporaryFile(mode='w+') as tmp:
      host.print_code(stencil, tmp)
      tmp.seek(0)
      if args.host_file == '-':
        shutil.copyfileobj(tmp, sys.stdout)
      else:
        with open(args.host_file, 'w') as host_file:
          shutil.copyfileobj(tmp, host_file)

  if args.header_file is not None:
    with tempfile.TemporaryFile(mode='w+') as tmp:
      header.print_code(stencil, tmp)
      tmp.seek(0)
      if args.header_file == '-':
        shutil.copyfileobj(tmp, sys.stdout)
      else:
        with open(args.header_file, 'w') as header_file:
          shutil.copyfileobj(tmp, header_file)

  if args.xo_file is not None:
    with tempfile.TemporaryFile(mode='w+b') as tmp_obj:
      rtl_kernel.print_code(
          stencil,
          tmp_obj,
          device_info=backend.parse_device_info(
              parser,
              args,
              platform_name='xocl_platform',
              part_num_name='xocl_part_num',
              clock_period_name='xocl_clock_period',
          ),
          rpt_file=args.xo_rpt,
          interface=args.interface,
      )
      tmp_obj.seek(0)
      if args.xo_file == '-':
        shutil.copyfileobj(tmp_obj, sys.stdout)   # type: ignore
      else:
        with open(args.xo_file, 'wb') as xo_file:
          shutil.copyfileobj(tmp_obj, xo_file)

  if args.output_dir is not None and (args.kernel_file is None or
                                      args.host_file is None or
                                      args.header_file is None):
    if args.kernel_file is None:
      dram_in = args.dram_in if args.dram_in else '_'
      dram_out = args.dram_out if args.dram_out else '_'
      kernel_file_name = os.path.join(
          args.output_dir, '%s_kernel-tile%s-unroll%d-ddr%s.cpp' % (
              stencil.app_name,
              'x'.join('%d'%x for x in stencil.tile_size[:-1]),
              stencil.unroll_factor, dram_in + '-' + dram_out))
    else:
      kernel_file_name = args.kernel_file
    with tempfile.TemporaryFile(mode='w+') as tmp:
      kernel.print_code(stencil, tmp)
      tmp.seek(0)
      with open(kernel_file_name, 'w') as kernel_file:
        shutil.copyfileobj(tmp, kernel_file)
    if args.host_file is None:
      host_file_name = os.path.join(args.output_dir, stencil.app_name + '.cpp')
    else:
      host_file_name = args.host_file
    with tempfile.TemporaryFile(mode='w+') as tmp:
      host.print_code(stencil, tmp)
      tmp.seek(0)
      with open(host_file_name, 'w') as host_file:
        shutil.copyfileobj(tmp, kernel_file)
    if args.header_file is None:
      header_file_name = os.path.join(args.output_dir, stencil.app_name + '.h')
    else:
      header_file_name = args.header_file
    with tempfile.TemporaryFile(mode='w+') as tmp:
      header.print_code(stencil, tmp)
      tmp.seek(0)
      with open(header_file_name, 'w') as header_file:
        shutil.copyfileobj(tmp, header_file)
